;; astronomy functions from "Practical Astronomy with Your 
;; Calculator" by Peter Duffett-Smith

;; note:  i'm aware that this is pretty gross lisp code, but it's
;; written to be easily matched with the algorithm descriptions
;; in the book -- so don't bug me about it!  :-)

;; changes:
;; 1/24/97:  added kepler's equation solver for better orbital
;;           accuracy (sections 47 and 61)
;; 2/3/98:   added orbital perturbations for jupiter and saturn (section 56)

;; temporary ease of use functions

(defun ll () (or (load 'astro2) (load "g:\\geostuff\\lspstuff\\astro2.lsp")))

;; macros, defstructs, globals:

(defmacro while (test &rest bod)
  `(do ()
       ((not ,test) nil)
     ,@bod))

;; astro-params definition and other "methods"

(defstruct astro-params
  (year 1998.0)		      ; local date
  (month 3.0)
  (day 1.0)
  (hour 12.0)		      ; local time
  (minute 0.0)
  (second 0.0)
  (tz -6)		      ; negative hours west of greenwich
  (dst nil)		      ; t if daylight savings time
  (longitude 93.15)	      ; location (my house)
  (latitude 44.9436))

(defun astro-params-lst (params)
  ;; local sidereal time
  (let* ((cd (julian-date-to-calendar-date
	      (local-calendar-date-and-time-to-julian-date
	       (astro-params-year params)
	       (astro-params-month params)
	       (astro-params-day params)
	       (astro-params-hour params)
	       (astro-params-minute params)
	       (astro-params-second params)
	       (astro-params-tz params)
	       (astro-params-dst params))))
	 (hms (dec-hms (* 24.0 (frac-part (third cd)))))
	 (gst (apply #'ut-gst
		     `(,(first cd) ,(second cd) ,(int-part (third cd)) ,@hms)))
	 (lst (apply #'gst-lst
		     (append gst (list (astro-params-longitude params))))))
    (apply #'hms-dec lst)))

(defun astro-params-dse (params)
  ;; days since epoch
  (julian-date-to-days-since-epoch
   (local-calendar-date-and-time-to-julian-date
    (astro-params-year params)
    (astro-params-month params)
    (astro-params-day params)
    (astro-params-hour params)
    (astro-params-minute params)
    (astro-params-second params)
    (astro-params-tz params)
    (astro-params-dst params))))

(defun astro-params-jd (params)
  ;; julian date
  (local-calendar-date-and-time-to-julian-date
   (astro-params-year params)
   (astro-params-month params)
   (astro-params-day params)
   (astro-params-hour params)
   (astro-params-minute params)
   (astro-params-second params)
   (astro-params-tz params)
   (astro-params-dst params)))

(if (not (boundp '*current-params*))	; kind of like defvar :-}
    (setq *current-params* (make-astro-params)))

(setq *key-west*
      (make-astro-params
       :year 1998
       :month 1
       :day 17
       :hour 12
       :minute 0
       :second 0
       :dst nil
       :longitude 81.7533
       :latitude 24.5722))

;; constants:

(setq *pi* 3.141592653589793238)

;; user interface:

(defun astro-help ()
  (format t "(set-params [:location <astro-params struct>])~%")
  (format t "(set-date-params) (set-time-params) (now)~%")
  (format t "(set-location-params)~%")
  (format t "(set-dst-param <t|nil>) (set-tz-param -6) ;; for cst~%")
  (format t "(print-params) (save-params)~%")
  (format t
    "(where [:sun | :moon | :mercury | :mars | ... ]) (whereall)~%")
  (format t "(current-horizon-coords body)  ;; same as (where body)~%")
  (format t "(current-ecliptic-coords body)~%")
  (format t
    "(current-equatorial-coords body) (current-equatorial-coords-h body)~%")
  (format t "(mapcar #'dec-dms *)  ")
  (format t ";; changes last value to deg min sec (or hour min sec)~%")
  nil)


(defun set-params (&key location)
  ;; let's make these globals elements of a class (well, defstruct)
  ;; among other things, this will allow us to keep a couple of contexts
  ;; around
  (if location (setf *current-params* (copy-astro-params location))
    (let (temp)
      (set-date-params)
      (set-time-params)
      (princ "Time zone (e.g. cst or number cst=-6):  ")
      (setq temp (read))
      (setf (astro-params-tz *current-params*)
	    (case temp ('est -5) ('cst -6) ('mst -7) ('pst -8) (t temp)))
      (princ "Daylight savings time in effect? (y, n):  ")
      (setf (astro-params-dst *current-params*) (read))
      (setf (astro-params-dst *current-params*)
	    (case (astro-params-dst *current-params*) ('y t) ('n nil)))
      (set-location-params)))
  'done)

(defun set-date-params ()
  (princ "Year (e.g. 1998):  ")
  (setf (astro-params-year *current-params*) (read))
  (princ "Month (1-12):  ")
  (setf (astro-params-month *current-params*) (read))
  (princ "day (1-31):  ")
  (setf (astro-params-day *current-params*) (read)))

(defun set-dst-param (val)
  ;; set daylight savings time t or nil
  (setf (astro-params-dst *current-params*) val))

(defun set-tz-param (n)
  ;; set timezone parameter, e.g. cst = -6
  (setf (astro-params-tz *current-params*) n))

(defun set-time-params ()
  (princ "Local time (hours):  ")
  (setf (astro-params-hour *current-params*) (read))
  (princ "Local time (minutes):  ")
  (setf (astro-params-minute *current-params*) (read))
  (princ "Local time (seconds):  ")
  (setf (astro-params-second *current-params*) (read)))

(defun set-location-params ()
  (let (deg min sec)
    (princ "Longitude (degrees):  ")
    (setf deg (read))
    (cond ((floatp deg) (setf min 0.0 sec 0.0))
	  (t (princ "Longitude (minutes or nil):  ")
	     (setf min (read))
	     (cond ((null min) (setf min 0.0 sec 0.0))
		   ((floatp min) (setf sec 0.0))
		   (t (princ "Longitude (seconds or nil):  ")
		      (setf sec (read))
		      (if (null sec) (setf sec 0.0))))))
    (setf (astro-params-longitude *current-params*) (dms-dec deg min sec))
    (princ "Latitude (degrees):  ")
    (setf deg (read))
    (cond ((floatp deg) (setf min 0.0 sec 0.0))
	  (t (princ "Latitude (minutes or nil):  ")
	     (setf min (read))
	     (cond ((null min) (setf min 0.0 sec 0.0))
		   ((floatp min) (setf sec 0.0))
		   (t (princ "Latitude (seconds or nil):  ")
		      (setf sec (read))
		      (if (null sec) (setf sec 0.0))))))
    (setf (astro-params-latitude *current-params*) (dms-dec deg min sec)))
  nil)

(defun now ()
  ;; set current time and date params to now
  (let* ((time (time))
         (date (date))
         (second (rem time 100))
         (minute (rem (/ time 100) 100))
         (hour (/ time 10000))
         (day (rem date 100))
         (month (rem (/ date 100) 100))
         (year (/ date 10000)))
    (setf (astro-params-year *current-params*) year)
    (setf (astro-params-month *current-params*) month)
    (setf (astro-params-day *current-params*) day)
    (setf (astro-params-hour *current-params*) hour)
    (setf (astro-params-minute *current-params*) minute)
    (setf (astro-params-second *current-params*) second)
    nil))

(defun print-params ()
  (format t "Year = ~a~%" (astro-params-year *current-params*))
  (format t "Month = ~a~%" (astro-params-month *current-params*))
  (format t "Day = ~a~%" (astro-params-day *current-params*))
  (format t "Hour = ~a~%" (astro-params-hour *current-params*))
  (format t "Minute = ~a~%" (astro-params-minute *current-params*))
  (format t "Second = ~a~%" (astro-params-second *current-params*))
  (format t "Time zone = ~a~%" (astro-params-tz *current-params*))
  (format t "Daylight savings time = ~a~%" (astro-params-dst *current-params*))
  (format t "Longitude = ~a~%" (astro-params-longitude *current-params*))
  (format t "Latitude = ~a~%" (astro-params-latitude *current-params*)))

(defun save-params ()
  (copy-astro-params *current-params*))

(defun whereall ()
  (mapc #'(lambda (planet)
	    (let ((coords (where planet)))
	      (format t "~a:  azimuth = ~a, altitude = ~a~%"
		      planet
		      (first coords)
		      (second coords))))
	'(:sun :moon :mercury :venus :mars
	  :jupiter :saturn :uranus :neptune :pluto))
  nil)

(defun where (body)
  (current-horizon-coords body))

(defun current-horizon-coords (body)
  (horizon-coords body
		  (astro-params-year *current-params*)
		  (astro-params-month *current-params*)
		  (astro-params-day *current-params*)
		  (astro-params-hour *current-params*)
		  (astro-params-minute *current-params*)
		  (astro-params-second *current-params*)
		  (astro-params-longitude *current-params*)
		  (astro-params-latitude *current-params*)
		  (astro-params-tz *current-params*)
		  (astro-params-dst *current-params*)))

(defun horizon-coords (body yy mm dd h m s long lat timezone dst-p)
  (let* ((jd (local-calendar-date-and-time-to-julian-date
	      yy mm dd h m s timezone dst-p)) ; preserve the date
	 (dse (julian-date-to-days-since-epoch jd))
	 (coords (ecliptic-coords body dse))
	 (date (julian-date-to-calendar-date jd))
	 (ut (* 24.0 (frac-part (third date))))
	 (time (dec-hms ut))
	 (gst (ut-gst (first date) (second date) (third date)
		      (first time) (second time) (third time)))
	 (lst (gst-lst (first gst) (second gst) (third gst) long)))
    (ecliptic-horizon (first coords) (second coords) lat
		      (apply #'hms-dec lst) yy)))

(defun ecliptic-coords (body dse)
  (case body
	(:sun (sun-coords dse))
	(:moon (moon-coords dse))
	(t (planet-coords body dse))))

(defun current-ecliptic-coords (body)
  (ecliptic-coords body (astro-params-dse *current-params*)))

(defun equatorial-coords (body dse latitude local-sidereal-time year)
  (let ((ecl (ecliptic-coords body dse)))
    (ecliptic-equatorial
     (first ecl) (second ecl) latitude local-sidereal-time year)))

(defun equatorial-coords-h (body dse latitude local-sidereal-time year)
  (let ((ec (equatorial-coords body dse latitude local-sidereal-time year)))
    (cons (/ (car ec) 15.0) (cdr ec))))

(defun current-equatorial-coords (body)
  (equatorial-coords
    body
    (astro-params-dse *current-params*)
    (astro-params-latitude *current-params*)
    (astro-params-lst *current-params*)
    (astro-params-year *current-params*)))

(defun current-equatorial-coords-h (body)
  (let ((ce (current-equatorial-coords body)))
    (cons (/ (car ce) 15.0) (cdr ce))))

;; the algorithms:

;; SECTION 4:  julian day numbers

(defun calendar-date-to-julian-date (yy mm dd &aux mmp yyp a b c d)
  ;; julian date corresponding to yy/mm/dd
  ;; dd can have a fractional part - the fraction of the day
  ;; e.g. 6:00 pm on february 17, 1985 yields
  ;; yy=1985, mm=2, and dd=17.75
  (if (or (= mm 1) (= mm 2))
    (setq yyp (1- yy) mmp (+ 12 mm))
    (setq yyp yy mmp mm))
  (if (or (>= yy 1582)
	  (and (= yy 1582) (>= mm 10))
	  (and (= yy 1582) (= mm 10) (>= dd 15)))
    (setq a (int-part (/ yyp 100)) b (+ 2 (- a) (int-part (/ a 4.0))))
    (setq b 0))
  (if (< yyp 0)
    (setq c (int-part (- (* 365.25 yyp) 0.75)))
    (setq c (int-part (* 365.25 yyp))))
  (setq d (int-part (* 30.6001 (1+ mmp))))
  (+ b c d dd 1720994.5))

(defun calendar-date-and-time-to-julian-date (yy mm dd h m s)
  (calendar-date-to-julian-date
   yy mm (+ (int-part dd)	; shouldn't have a fractional day
	    (/ (hms-dec h m s) 24.0))))

(defun local-calendar-date-and-time-to-julian-date
       (yy mm dd h m s timezone dst-p)
  ;; timezone is negative to the west - cst is -6
  ;; dst-p is true if daylight savings time is in effect
  (- (calendar-date-and-time-to-julian-date yy mm dd h m s)
     (/ (if dst-p 1.0 0.0) 24.0)
     (/ timezone 24.0)))

(defun julian-date-at-0-hours (jd)
  (+ 0.5 (int-part (- jd 0.5))))

;; days since the epoch 1990 January 0.0

(defun calendar-date-to-days-since-epoch (yy mm dd)
  ;; days since the epoch 1990 January 0.0
  (julian-date-to-days-since-epoch (calendar-date-to-julian-date yy mm dd)))

(defun julian-date-to-days-since-epoch (jd)
  ;; converts julian date to days since epoch 1990 january 0.0
  (- jd 2447891.5))

(defun days-since-epoch-to-julian-date (dse)
  ;; the reverse
  (+ dse 2447891.5))

;; SECTION 5:  converting the julian day number to the calendar date

(defun julian-date-to-calendar-date (jd)
  (let* ((jd (+ jd 0.5))
	 (i (int-part jd))
	 (f (frac-part jd))
	 (a (int-part (/ (- i 1867216.25) 36524.25)))
	 (b (if (> i 2299160)
	      (+ i 1 a (- (int-part (/ a 4.0))))
	      i))
	 (c (+ b 1524))
	 (d (int-part (/ (- c 122.1) 365.25)))
	 (e (int-part (* 365.25 d)))
	 (g (int-part (/ (- c e) 30.6001)))
	 (dd (+ c (- e) f (- (int-part (* 30.6001 g)))))
	 (mm (if (< g 13.5) (- g 1) (- g 13)))
	 (yy (if (> mm 2.5) (- d 4716) (- d 4715))))
    (list yy mm dd)))

;; SECTION 6:  finding the day of the week

(defun dow (jd)
  ;; returns day of the week - sunday = 0, ... , saturday = 6
  (let* ((jd (julian-date-at-0-hours jd))
	 (a (/ (+ jd 1.5) 7.0)))
    (round (* 7 (frac-part a)))))

;; SECTION 7:  converting hours, minutes, and seconds to decimal hours

(defun hms-dec (hh mm ss)
  ;; convert hours, minutes, seconds to decimal hours
  (+ hh (/ (+ mm (/ ss 60.0)) 60.0)))

;; also works for degrees:

(defun dms-dec (deg min sec)
  ;; convert degrees, minutes, seconds to decimal degrees
  (hms-dec deg min sec))

;; SECTION 8: converting decimal hours to hours, minutes, and seconds

(defun dec-hms (hours)
  ;; convert decimal hours to hours, minutes, seconds
  (let* ((hh (int-part hours))
         (tmp1 (* 60 (frac-part hours)))
         (mm (int-part tmp1))
         (ss (* 60 (frac-part tmp1))))
    (list hh mm ss)))

(defun dec-dms (degrees)
  ;; convert decimal degrees to degrees, minutes, seconds
  (dec-hms degrees))

;; SECTION 9:  converting local time to UT

(defun lt-ut (lt)
  ;; convert local standard time to universal time
  (reduce-to-24 (- lt (astro-params-tz *current-params*))))

;; SECTION 10:  converting UT to local civil time

(defun ut-lt (ut)
  ;; convert universal time to local standard time
  (reduce-to-24 (+ ut (astro-params-tz *current-params*))))

;; SECTION 12:  conversion of UT to GST

;; gdh - we should redo these so that they're all done in decimal!!!

(defun ut-gst (yy mm dd h m s)
  ;; convert universal time to greenwich mean sidereal time
  (let* ((jd (calendar-date-to-julian-date yy mm (int-part dd)))
         (ss (- jd 2451545.0))
         (tt (/ ss 36525.0))
         (t0 (reduce-to-24
               (+ 6.697374558
                  (* 2400.051336 tt)
                  (* 0.000025862 tt tt))))
         (ut (* (hms-dec h m s) 1.002737909))
	 (gst (reduce-to-24 (+ t0 ut))))
    (dec-hms gst)))

;; SECTION 13:  conversion of GST to UT

(defun gst-ut (yy mm dd h m s)
  ;; convert greenwich mean sidereal time to universal time
  (let* ((jd (calendar-date-to-julian-date yy mm (int-part dd)))
	 (ss (- jd 2451545.0))
	 (tt (/ ss 36525.0))
	 (t0 (reduce-to-24
               (+ 6.697374558
                  (* 2400.051336 tt)
                  (* 0.000025862 tt tt))))
	 (gst (hms-dec h m s))
	 (tmp (reduce-to-24 (- gst t0)))
	 (ut (* tmp 0.9972695663)))
    (dec-hms ut)))

;; SECTION 14:  local sidereal time (LST)
	 
(defun gst-lst (h m s longitude)
  ;;convert greenwich sidereal time to local sidereal time
  ;;west longitudes are positive
  (let* ((gst (hms-dec h m s))
	 (timedif (/ longitude 15.0))
	 (lst (reduce-to-24 (- gst timedif))))
    (dec-hms lst)))

;; SECTION 15:  converting LST to GST

(defun lst-gst (h m s longitude)
  ;;convert local sidereal time to greenwich sidereal time
  ;;west longitudes are positive
  (let* ((lst (hms-dec h m s))
	 (timedif (/ longitude 15.0))
	 (lst (reduce-to-24 (+ lst timedif))))
    (dec-hms lst)))

;; SECTION 27:  ecliptic to equatorial coordinates conversion

;; well, actually, this is where the obliquity of the ecliptic is defined,
;; and that's what we'll do

(defun ecliptic-obliquity (year)
  ;; obliquity of the ecliptic (epsilon) at year January 0.0
  (let* ((jd (- (calendar-date-to-julian-date year 1 0.0)
		2451545.0))		; jd for 2000 January 1.5
	 (tt (/ jd 36525.0))		; julian centuries
	 (deleps (/ (+ (* 46.815 tt) (* .0006 tt tt) (* .00181 tt tt tt))
		    3600.0)))
    (- 23.439292 deleps)))

;(defun ecliptic-obliquity ()
;  ;; in degrees (p 49)
;  23.446236)

;; SECTION 31:  generalised coordinate transformations

(defun ecliptic-equatorial (lambda beta latitude local-sidereal-time year)
  ;; returns right ascension, declination
  (vector-coords
    (matmult
      (conversion-matrix :lambda-beta-to-alpha-delta
			 latitude
			 local-sidereal-time
			 (ecliptic-obliquity year))
      (coords-vector lambda beta))))

(defun ecliptic-horizon (lambda beta latitude local-sidereal-time year)
  ;; returns azimuth, altitude
  (vector-coords
    (matmult
      (conversion-matrix :h-delta-to-a-a
			 latitude
			 local-sidereal-time
			 (ecliptic-obliquity year))
      (conversion-matrix :alpha-delta-to-h-delta
			 latitude
			 local-sidereal-time
			 (ecliptic-obliquity year))
      (conversion-matrix :lambda-beta-to-alpha-delta
			 latitude
			 local-sidereal-time
			 (ecliptic-obliquity year))
      (coords-vector lambda beta))))

(defun coords-vector (u v)
  ;; convert coordinates to a 1x3 column vector
  ;; input in degrees
  (setq u (deg-rad u) v (deg-rad v))
  (make-matrix 1 3 (* (cos u) (cos v)) (* (sin u) (cos v)) (sin v)))

(defun vector-coords (vec)
  ;; convert a 1x3 column vector to coordinates
  (list
   (dunambiguous-atan
    (get-matrix-entry vec 0 1)
    (get-matrix-entry vec 0 0))
   (dasin (get-matrix-entry vec 0 2))))

(defun conversion-matrix (type latitude local-sidereal-time epsilon)
  ;; assume here that latitude and local-sidereal-time are decimal
  ;; also that local-sidereal-time is in hours
  ;; epsilon is the obliquity of the ecliptic in degrees
  ;;
  ;; coordinates guide:
  ;;  a-a:		azimuth, altitude			horizon
  ;;  h-delta:		hour-angle, declination			equatorial
  ;;  alpha-delta	right ascension, declination		equatorial
  ;;  lambda-beta	ecliptic longitude, ecliptic latitude	ecliptic
  ;;  l-b		galactic longitude, galactic latitude	galactic
  (setq latitude (deg-rad latitude)
	local-sidereal-time (deg-rad (* 15.0 local-sidereal-time))
	epsilon (deg-rad epsilon))
  (case type
     ((:a-a-to-h-delta :h-delta-to-a-a)	; matrix A
      (make-matrix 3 3
	 (- (sin latitude))	 0.0			(cos latitude)
	 0.0			-1.0			0.0
	 (cos latitude)		 0.0			(sin latitude)))
     ((:alpha-delta-to-h-delta :h-delta-to-alpha-delta)	; matrix B
      (make-matrix 3 3
	 (cos local-sidereal-time)	(sin local-sidereal-time)	0.0
	 (sin local-sidereal-time)	(- (cos local-sidereal-time))	0.0
	 0.0				0.0				1.0))
     (:alpha-delta-to-lambda-beta	; matrix C
      (make-matrix 3 3
	 1.0			 0.0			0.0
	 0.0			(cos epsilon)		(sin epsilon)
	 0.0			(- (sin epsilon))	(cos epsilon)))
     (:lambda-beta-to-alpha-delta	; matrix C-prime
      (make-matrix 3 3
	 1.0			 0.0			0.0
	 0.0			(cos epsilon)		(- (sin epsilon))
	 0.0			(sin epsilon)		(cos epsilon)))
     (:alpha-delta-to-l-b		; matrix D
      (make-matrix 3 3
	 -0.0669887	-0.8727558	-0.4835389
	  0.4927285	-0.4503470	 0.7445846
	 -0.8676008	-0.1883746	 0.4601998))
     (:l-b-to-alpha-delta		; matrix D-prime
      (make-matrix 3 3
	 -0.0669887	 0.4927285	-0.8676008
	 -0.8727558	-0.4503470	-0.1883746	 
	 -0.4835389	 0.7445846	 0.4601998))))

;;matrix functions

(defun make-matrix (w h &rest ivals)
  ;; make a new matrix and initialize it.  values are filled in in
  ;; row major order.  if values run out before matrix positions,
  ;; the last is repeated
  ;; default initial value is 0.0
  (or ivals (setf ivals (list 0.0)))
  (let ((v (make-array h)))
    (dotimes (i h)
       (setf (aref v i) (make-array w)))
    ;; initialize
    (dotimes (j h)
       (dotimes (i w)
	  (set-matrix-entry v i j (car ivals))
	  (if (cdr ivals) (setq ivals (cdr ivals)))))
    v))

(defun get-matrix-entry (matrix i j)
  (aref (aref matrix j) i))

(defun set-matrix-entry (matrix i j value)
  (setf (aref (aref matrix j) i) value))

(defun mwidth (matrix)
  (length (aref matrix 0)))

(defun mheight (matrix)
  (length matrix))

(defun print-matrix (matrix)
  (dotimes (j (mheight matrix))
    (dotimes (i (mwidth matrix))
	 (princ (aref (aref matrix j) i))
	 (princ " "))
    (terpri)))

(defun matmult (mat1 &rest matrices)
  (let ((result mat1))
    (dolist (mat matrices)
      (setf result (matmult-2arg result mat)))
    result))

(defun matmult-2arg (mat1 mat2)
  ;; multiplies matrices mat1 and mat2 together
  (if (not (eql (mwidth mat1) (mheight mat2)))
    (xlerror "mismatched matrices"))
  (let ((newmat (make-matrix (mwidth mat2) (mheight mat1))))
    (dotimes (j (mheight mat1))
      (dotimes (i (mwidth mat2))
        (dotimes (k (mwidth mat1))
          (set-matrix-entry newmat i j 
            (+ (get-matrix-entry newmat i j) 
               (* (get-matrix-entry mat1 k j)
                  (get-matrix-entry mat2 i k)))))))
    newmat))

;; SECTION 32:  the angle betwen two celestial objects

(defun celestial-angle-equatorial (alpha1 delta1 alpha2 delta2)
  ;; equatorial coordinates - right ascension, declination
  (let ((angle (dacos (+ (* (dsin delta1) (dsin delta2))
			 (* (dcos delta1) (dcos delta2) (dcos (- alpha1 alpha2)))))))
    (if (or (< (abs angle) 0.00290888)	; 10 arc minutes
	    (< (abs (- 180.0 angle)) 0.00290888))
      (celestial-angle-equatorial-small alpha1 delta1 alpha2 delta2)
      angle)))


(defun celestial-angle-ecliptic (lambda1 beta1 lambda2 beta2)
  ;; equatorial coordinates - ecliptic longitude, ecliptic latitude
  (celestial-angle-equatorial lambda1 beta1 lambda2 beta2))

(defun celestial-angle-equatorial-small (alpha1 delta1 alpha2 delta2)
  ;; equatorial coordinates - right ascension, declination
  ;; when angle is small
  (sqrt (+ (* (expt (dcos delta1) 2) (expt (- alpha2 alpha1) 2))
	   (expt (- delta2 delta1) 2))))

;; SECTION 33:  rising and setting

(defun object-rise-time (latitude alpha delta)
  ;; returns local sidereal time of object rise
  ;; equatorial coordinates - alpha = right ascension, delta = declination
  (let ((foo (- (* (dtan latitude) (dtan delta)))))
    (if (and (<= -1.0 foo) (<= foo 1.0))
      (reduce-to-24 (+ 24 (/ (dacos foo) 15.0) alpha))
      (format t "this object doesn't rise.~%"))))

(defun object-set-time (latitude alpha delta)
  ;; returns local sidereal time of object set
  ;; equatorial coordinates - alpha = right ascension, delta = declination
  (let ((foo (- (* (dtan latitude) (dtan delta)))))
    (if (and (<= -1.0 foo) (<= foo 1.0))
      (reduce-to-24 (+ (/ (dacos foo) 15.0) alpha))
      (format t "this object doesn't set.~%"))))

(defun object-rise-azimuth (latitude alpha delta)
  ;; returns azimuth of object rise
  ;; equatorial coordinates - alpha = right ascension, delta = declination
  (let ((foo (/ (dsin delta) (dcos latitude))))
    (if (and (<= -1.0 foo) (<= foo 1.0))
      (dacos foo)
      (format t "this object doesn't rise.~%"))))

(defun object-set-azimuth (latitude alpha delta)
  ;; returns azimuth of object set
  ;; equatorial coordinates - alpha = right ascension, delta = declination
  (let ((foo (/ (dsin delta) (dcos latitude))))
    (if (and (<= -1.0 foo) (<= foo 1.0))
      (- 360.0 (dacos foo))
      (format t "this object doesn't rise.~%"))))

;; have to add the refraction stuff now

;; SECTION 46:  calculating the position of the sun

(defun sun-coords (dse)
  ;; return a list of lambda and beta - note: beta is always 0.0
  (list (sun-longitude dse) 0.0))

(defun sun-longitude (dse)
  ;; calculate lambda sub sun, the longitude in ecliptic coordinates
  ;; dse is days since epoch and can be fractional
  ;; the sun's latitude is by definition 0 in ecliptic coordinates
  (let* ((ecliptic-longitude-at-epoch 279.403303) ; epsilon sub g
	 (ecliptic-longitude-of-perigee 282.768422) ; omega bar sub g
	 (eccentricity-of-orbit 0.016713) ; e
	 (mean-anomoly (reduce-to-360	; M sub sun
			(+ (/ (* 360.0 dse) 365.242191)
			   ecliptic-longitude-at-epoch
			   (- ecliptic-longitude-of-perigee))))
; less accurate way:
;	 (true-anomoly (reduce-to-360	; nu (or v)
;			(+ mean-anomoly
;			   (/ (* 360.0
;				 eccentricity-of-orbit
;				 (dsin mean-anomoly))
;			      *pi*))))
	 (true-anomoly (true-anomoly eccentricity-of-orbit mean-anomoly)))
    (reduce-to-360 (+ true-anomoly ecliptic-longitude-of-perigee))))

;; SECTION 47:  calculating orbits more precisely

(defun true-anomoly (eccentricity-of-orbit mean-anomoly)
  ;; returns degrees
  (* 2.0
     (datan
      (* (sqrt (/ (+ 1 eccentricity-of-orbit)
		  (- 1 eccentricity-of-orbit)))
	 (tan (/ (eccentric-anomoly eccentricity-of-orbit mean-anomoly)
		 2.0))))))
	 
;; solve kepler's equation:  E - e * sin(E)=M sub sun

(defun eccentric-anomoly (eccentricity-of-orbit mean-anomoly) ; e and M
  ;; mean-anomoly is in degrees; we convert it to radians for the calculations
  ;; result is in radians
  (do* ((mean-anomoly (deg-rad mean-anomoly))
	(eccentric-anomoly mean-anomoly) ; guess ok for e < ~ 0.1
	(required-accuracy 1e-6)	; epsilon
;	(n 0 (1+ n))
	(delta 0.0)
	)
      (nil)
    (setf delta (- eccentric-anomoly
		   (* eccentricity-of-orbit
		      (sin eccentric-anomoly)) ; radians
		   mean-anomoly))
;    (format t "~a:  eccentric anomoly = ~a, delta = ~a~%"
;	    n eccentric-anomoly delta)
    (if (< (abs delta) required-accuracy) (return eccentric-anomoly))
    (setf eccentric-anomoly (- eccentric-anomoly
			       (/ delta
				  (- 1 (* eccentricity-of-orbit
					  (cos eccentric-anomoly))))))))

;; SECTION 54:  calculating the coordinates of a planet

(setq *planet-data*
  ;; planet, orbital period, longitude at epoch, longitude at perihelion,
  ;; orbital eccentricity, semi-major axis of the orbit, inclination of
  ;; the orbit, longitude of the ascending node
      '((:mercury 0.240852 60.750646 77.299833 0.205633 0.387099
		  7.004540 48.212740)
	(:venus 0.615211 88.455855 131.430236 0.006778 0.723332
		3.394535 76.589820)
	(:earth 1.00004 99.403308 102.768413 0.016713 1.00000 nil nil)
	(:mars 1.880932 240.739474 335.874939 0.093396 1.523688
	       1.849736 49.480308)
	(:jupiter 11.863075 90.638185 14.170747 0.048482 5.202561
		  1.303613 100.353142)
	(:saturn 29.471362 287.690033 92.861407 0.055581 9.554747
		 2.488980 113.576139)
	(:uranus 84.039492 271.063148 172.884833 0.046321 19.21814
		 0.773059 73.926961)
	(:neptune 164.79246 282.349556 48.009758 0.009003 30.109570
		  1.770646 131.670599)
	(:pluto 246.77027 221.4127 224.133 0.24624 39.3414
		17.1420 110.144)))

(defun planet-coords-2 (planet dse)
  (let* ((coords (planet-coords planet dse))
	 (lambda (car coords))
	 (beta (cadr coords)))
    (list lambda (if (> beta 270) (- beta 360) beta))))

(defun planet-coords (planet dse)
  ;; calculate the ecliptic coordinates of a planet
  ;; planet is :mercury, :venus, :mars, ...
  ;; dse is days-since-epoch
  (let* ((planet-data (cdr (assoc planet *planet-data*)))
	 (landr-planet (heliocentric-longitude-and-radius planet dse))
	 (planet-l (first landr-planet))  ; l
	 (planet-r (second landr-planet)) ; r
	 (landr-earth (heliocentric-longitude-and-radius :earth dse))
	 (earth-l (first landr-earth))	; L
	 (earth-r (second landr-earth))	; R
	 (orbital-inclination (sixth planet-data)) ; i
	 (longitude-of-ascending-node (seventh planet-data)) ; omega
	 (heliocentric-latitude		; psi
	  (reduce-to-360
	   (dasin (* (dsin (- planet-l longitude-of-ascending-node))
		     (dsin orbital-inclination)))))
	 ;; now project onto the plane of the ecliptic
	 (projected-heliocentric-longitude ; l-prime
	  (reduce-to-360
	   (+ (dunambiguous-atan
	       (* (dsin (- planet-l longitude-of-ascending-node))
		  (dcos orbital-inclination))
	       (dcos (- planet-l longitude-of-ascending-node)))
	      longitude-of-ascending-node)))
	 (projected-radius-vector	; r-prime
	  (* planet-r (dcos heliocentric-latitude)))
	 ;; now refer calculations to the earth to get geocentric
	 ;; ecliptic latitude (beta) and longitude (lambda) of the planet
	 (ecliptic-longitude		; lambda
	  (case planet
		((:mercury :venus)	; the inner planets
		 (reduce-to-360
		  (+
		   180.0
		   earth-l
		   (datan
		    (/
		     (* projected-radius-vector
			(dsin (- earth-l projected-heliocentric-longitude)))
		     (- earth-r
			(* projected-radius-vector
			   (dcos
			    (- earth-l
			       projected-heliocentric-longitude)))))))))
		((:mars :jupiter :saturn :uranus :neptune :pluto)
		 ;; the outer planets
		 (reduce-to-360
		  (+
		   (datan
		    (/
		     (* earth-r
			(dsin (- projected-heliocentric-longitude earth-l)))
		     (- projected-radius-vector
			(* earth-r
			   (dcos
			    (- projected-heliocentric-longitude earth-l))))))
		   projected-heliocentric-longitude)))))
	 (ecliptic-latitude		; beta
	  (reduce-to-360
	   (datan
	    (/
	     (* projected-radius-vector
		(dtan heliocentric-latitude)
		(dsin (- ecliptic-longitude projected-heliocentric-longitude)))
	     (* earth-r
		(dsin (- projected-heliocentric-longitude earth-l))))))))
    (list ecliptic-longitude ecliptic-latitude)))

(defun heliocentric-longitude-and-radius (planet dse)
  ;; calculate the heliocentric longitude and radius of a planet
  ;; planet is :mercury, :venus, :mars, ...
  ;; dse is days-since-epoch
  (let* ((planet-data (cdr (assoc planet *planet-data*)))
	 (orbital-period (first planet-data)) ; Tsubp
	 (longitude-at-epoch (second planet-data)) ; epsilon
	 (longitude-at-perihelion (third planet-data)) ; small omega bar
	 (orbital-eccentricity (fourth planet-data)) ; e
	 (semi-major-orbit-axis (fifth planet-data)) ; a
	 (mean-anomoly (reduce-to-360	; M
			(+ (/ (* 360.0 dse) (* 365.242191 orbital-period))
			   longitude-at-epoch
			   (- longitude-at-perihelion))))
; less accurate way:
;	 (true-anomoly (reduce-to-360	; v
;			(+ mean-anomoly
;			   (/ (* 360.0
;				 orbital-eccentricity
;				 (dsin mean-anomoly))
;			      *pi*))))
	 (true-anomoly (true-anomoly orbital-eccentricity mean-anomoly))
	 (heliocentric-longitude (reduce-to-360	; l
				  (+ true-anomoly
				     longitude-at-perihelion
				     (perturbation-delta planet dse))))
	 (radius (/ (* semi-major-orbit-axis ; r
		       (- 1.0 (* orbital-eccentricity orbital-eccentricity)))
		    (+ 1.0
		       (* orbital-eccentricity
			  (dcos true-anomoly))))))
    (list heliocentric-longitude radius)))

;; SECTION 56:  perturbatons in a planet's orbit

(defun perturbation-delta (planet dse)
  (if (not (or (eq planet :jupiter) (eq planet :saturn)))
    0.0
    (let* ((jd (days-since-epoch-to-julian-date dse))
	   (tt (/ (- jd 2415020.0)
		  36525.0))		; julian centuries since 1900 jan 0.5
	   (a (+ (/ tt 5.0) 0.1))
	   (p (+ 237.47555 (* 3034.9061 tt)))
	   (q (+ 265.91650 (* 1222.1139 tt)))
	   (v (- (* 5.0 q) (* 2.0 p)))
	   (b (- q p)))
      (if (eq planet :jupiter)
	(- (* (- 0.3314 (* 0.0103 a))
	      (dsin v))
	   (* 0.0644 a (dcos v)))
	;; else :planet is :saturn
	(+ (* (- (* 0.1609 a) 0.0105) (dcos v))
	   (* (- (* 0.0182 a) 0.8142) (dsin v))
	   (- (* 0.1488 (dsin b)))
	   (- (* 0.0408 (dsin (* 2.0 b))))
	   (* 0.0856 (dsin b) (dcos q))
	   (* 0.0813 (dcos b) (dsin q)))))))

;; SECTION 65:  calculating the moon's position

(defun moon-coords (dse)
  (let* ((sun-ecliptic-longitude-at-epoch 279.403303)
	 (sun-ecliptic-longitude-of-perigee 282.768422)
	 (sun-eccentricity-of-orbit 0.016713)
	 (sun-mean-anomoly (reduce-to-360
			    (+ (/ (* 360.0 dse) 365.242191)
			       sun-ecliptic-longitude-at-epoch
			       (- sun-ecliptic-longitude-of-perigee))))
	 (sun-true-anomoly (reduce-to-360
			    (+ sun-mean-anomoly
			       (/ (* 360.0
				     sun-eccentricity-of-orbit
				     (dsin sun-mean-anomoly))
				  *pi*))))
	 (sun-longitude (reduce-to-360
			 (+ sun-true-anomoly
			    sun-ecliptic-longitude-of-perigee)))
	 (moon-mean-longitude-at-epoch 318.351648) ; lsub0
	 (moon-mean-longitude-of-perigee-at-epoch 36.340410) ; Psub0
	 (moon-mean-longitude-at-node-at-epoch 318.510107) ; Nsub0
	 (moon-inclination-of-orbit 5.145396) ; i
	 (moon-eccentricity-of-orbit 0.054900) ; e
	 (moon-semi-major-axis-of-orbit 384401)	; a
	 (moon-mean-longitude		; l
	  (+ (* 13.1763966 dse) moon-mean-longitude-at-epoch))
	 (moon-mean-anomoly		; Msubm
	  (- moon-mean-longitude
	     (* 0.1114041 dse)
	     moon-mean-longitude-of-perigee-at-epoch))
	 (moon-ascending-node-mean-longitude ; N
	  (- moon-mean-longitude-at-node-at-epoch (* 0.052953 dse)))
	 (evection-correction		; Esubv
	  (* 1.2739
	     (dsin (- (* 2.0 (- moon-mean-longitude sun-longitude))
		      moon-mean-anomoly))))
	 (annual-equation-correction	; Asube
	  (* 0.1858 (dsin sun-mean-anomoly)))
	 (third-correction (* 0.37 (dsin sun-mean-anomoly))) ; Asub3
	 (moon-corrected-anomoly	; M-prime-subm
	  (+ moon-mean-anomoly
	     evection-correction
	     (- annual-equation-correction)
	     (- third-correction)))
	 (equation-of-the-center-correction ; Esubc
	  (* 6.2886 (dsin moon-corrected-anomoly)))
	 (fourth-correction		; Asub4
	  (* 0.214 (dsin (* 2.0 moon-corrected-anomoly))))
	 (moon-corrected-longitude	; l-prime
	  (+ moon-mean-longitude
	     evection-correction
	     equation-of-the-center-correction
	     (- annual-equation-correction)
	     fourth-correction))
	 (variation-correction		; V
	  (* 0.6583 (dsin (* 2.0 (- moon-corrected-longitude
				    sun-longitude)))))
	 (moon-true-orbital-longitude	; l-prime-prime
	  (+ moon-corrected-longitude variation-correction))
	 (corrected-longitude-of-the-node ; N-prime
	  (- moon-ascending-node-mean-longitude
	     (* 0.16 (dsin sun-mean-anomoly))))
	 (ecliptic-longitude		; lambdasubm
	  (reduce-to-360
	   (+ (dunambiguous-atan
	       (* (dsin (- moon-true-orbital-longitude
			   corrected-longitude-of-the-node))
		  (dcos moon-inclination-of-orbit))
	       (dcos (- moon-true-orbital-longitude
			corrected-longitude-of-the-node)))
	      corrected-longitude-of-the-node)))
	 (ecliptic-latitude		; betasubm
	  (reduce-to-360
	   (dasin (* (dsin (- moon-true-orbital-longitude
			      corrected-longitude-of-the-node))
		     (dsin moon-inclination-of-orbit))))))
    (list ecliptic-longitude ecliptic-latitude)))

;; helpers

(defun unambiguous-atan (y x)
  ;; arctan of y/x
  (cond ((zerop x) (if (<= y 0) (* 3.0 (/ *pi* 2)) (/ *pi* 2)))
	((< x 0.0) (+ *pi* (atan (/ y x))))
	((< y 0.0) (+ (* 2 *pi*) (atan (/ y x))))
	(t (atan (/ y x)))))

(defun dsin (x)
  (sin (deg-rad x)))

(defun dcos (x)
  (cos (deg-rad x)))

(defun dtan (x)
  (tan (deg-rad x)))

(defun dasin (x)
  (rad-deg (asin x)))

(defun dacos (x)
  (rad-deg (acos x)))

(defun datan (x)
  (rad-deg (atan x)))

(defun dunambiguous-atan (y x)
  (rad-deg (unambiguous-atan y x)))

(defun reduce-to-n (x n)
  ;; add or subtract multiples of n to bring into range
  ;; [0, n)  --  like "mod", but with real numbers
  (let ((incr (if (< x 0) n (- n))))
    (while (or (< x 0) (>= x n))
      (setq x (+ x incr))))
  x)

(defun reduce-to-24 (x)
  ;; add or subtract multiples of 24 to bring into range
  ;; [0,24)
  (reduce-to-n x 24.0))

(defun reduce-to-360 (x) (reduce-to-n x 360.0))

(defun int-part (x)
  (truncate x))

(defun frac-part (x)
  (- x (truncate x)))

(defun deg-rad (x)
  (/ (* x *pi*) 180.0))

(defun rad-deg (x)
  (* x (/ 180.0 *pi*)))

(defun fifth (l)
  (car (cddddr l)))

(defun sixth (l)
  (cadr (cddddr l)))

(defun seventh (l)
  (caddr (cddddr l)))

(defun round (x)
  (truncate (+ x 0.5)))

;; message to the user:

(format t "Type \"(astro-help)\" for help.~%")
